-- TARGET LOG!
-- Tracks mouseover targets!
-- Now with Grudge log!

--==========================================================================--
-- Global constants

BINDING_HEADER_TARGETLOG = "TargetLog Buttons";
BINDING_NAME_TARGETLOGTOGGLE = "Toggle TargetLog window";

--==========================================================================--
-- Local constants

TLOG_VER_MAJOR = "0";
TLOG_VER_MINOR = "9";
TLOG_VER_REVIS = "7";

TLOG_LOG_SIZE = 16;
TLOG_MAX_LOG_SHOW = 16;
TLOG_DEFAULT_ALPHA = 1.0;

GRUDGEFRAME_HEIGHT = 16;
GRUDGEFRAME_DISPLAYNUM = 17;

GRUDGESHARE_HEADER = "TLGrudgeShare";

TLOG_DOUBLECLICK_TIME	= 0.33;	-- How long within you have to double-click to have it count 
TLOG_ONUPDATE_TIME	= 2.0;	-- How often target log window re-updates to account for non-mouseover target changes
TLOG_FIELDUPDATE_TIME	= 1.5;	-- How often the description field changes
TLOG_FIELDFADETIME	= 0.75;	-- Fade time

TLOG_NO_LEVEL		= -32767;

-- Affiliation constants
ENEMY_PC	= "EnemyPC";
ENEMY_NPC	= "EnemyNPC";
FRIENDLY_PC	= "FriendlyPC";
FRIENDLY_NPC	= "FriendlyNPC";
NEUTRAL_NPC	= "Neutral";
FRIENDLY_PARTY	= "Party";
FRIENDLY_GUILD	= "Guild";
UNKNOWN_AFF	= "Unknown";

-- TargetLog colours:
DEFCOLOUR = "|cffafafaf";	-- colour of description text
HICOLOUR  = "|cffffffff";

-- GrudgeList colours:
GRUDGEWARNCOLOUR = "|cffff0000";	-- colour of popup warning
GRUDGELISTCOLOUR = "|cffffff00";	-- colour of chat screen list text
GRUDGEFRAMECOLOUR = "|cffffffff";	-- colour of dialog box text

-- Various name colours
ENEMY_PC_COLOUR		= "|cffff0000";
FRIEND_PC_COLOUR	= "|cff00ff00";

ENEMY_NPC_COLOUR	= "|cffaf0000";
FRIEND_NPC_COLOUR	= "|cff007f00";

NEUTRALCOLOUR		= "|cffffff00";
PARTYCOLOUR		= "|cff4f7fff";
GUILDCOLOUR		= "|cff0000ff";

ELITECOLOUR		= "|cffffff4f";

-- Vertex colours for TargetLog highlights
VERT_NORM_R = 1.0;
VERT_NORM_G = 1.0;
VERT_NORM_B = 0.0;

VERT_TARG_R = 0.5;
VERT_TARG_G = 0.5;
VERT_TARG_B = 0.5;

VERT_WARN_R = 1.0;
VERT_WARN_G = 0.0;
VERT_WARN_B = 0.0;

DEFAULT_SORT_ORDER = {"name","level","race","class","guild"};

--==========================================================================--
-- Global variables

UIPanelWindows["TargetLogGrudgeFrame"] = { areas = "left", pushable = 5 , whileDead = 1};

--==========================================================================--
-- Local variables

local TotalGrudges = 0;
local CurrentGrudge = "";
local GrudgePanelSelect = nil;

local CurrentSortType = "name";
local SortedGrudgeList = { };
local SortRevers = nil;

local RealmName = "";

local TemporaryDisable = nil;

local DoubleClick = {-1, 0.0};

local IsEditingReason = nil;

local TemporaryGrudgeList = { };

local FieldDisplay = "class";
local LastUpdateTime = TLOG_ONUPDATE_TIME;
local LastFieldUpdateTime = TLOG_FIELDUPDATE_TIME;

local LastTargetName = "";

--==========================================================================--
-- etc...

function TargetLog_Event()
	if (event == "VARIABLES_LOADED") then
		TargetLog_DataLoaded();
	elseif (event == "UPDATE_MOUSEOVER_UNIT") then
		TargetLog_UpdateList("mouseover");
		TargetLogHistory_Update();
		TargetLogGrudge_UpdateButtonStates()
	elseif (event == "PLAYER_TARGET_CHANGED") then
	--elseif ((event == "PLAYER_TARGET_CHANGED") or (event == "UNIT_HEALTH")) then
		-- Bugged, target change doesn't always trigger properly?
		--DEFAULT_CHAT_FRAME:AddMessage("Player target changed");
		LastTargetName = UnitName("target");
		TargetLog_UpdateList("target");
		TargetLogHistory_Update();
		TargetLogGrudge_UpdateButtonStates()
	elseif (event == "CHAT_MSG_GUILD") then
		if (TargetLogState.GuildShare) then
			TargetLog_ParseGuildMsg(arg1, arg2);
		end
	end
end

function TargetLog_Init()
	this:RegisterForDrag("LeftButton");
	local Loaded = "";

	-- Get current realm name
	RealmName = string.gsub(GetCVar("realmName"),"'","");

	-- Variables saved
	--RegisterForSave("TargetLogData");
	--RegisterForSave("TargetLogState");
	--RegisterForSave("TargetLogGrudge");

	-- Events to listen for
	this:RegisterEvent("VARIABLES_LOADED");
	this:RegisterEvent("UPDATE_MOUSEOVER_UNIT");
	this:RegisterEvent("UNIT_HEALTH");
	this:RegisterEvent("CHAT_MSG_GUILD");

	-- TODO: Some other events which we may consider in the future:
	-- Combat messages involving grudged players, we can listen for to
	-- detect their proximity
	-- Also emote messages

	-- Register commands
	SLASH_TLOG1="/tlog";
	SLASH_TLOG2="/targetlog";
	SlashCmdList["TLOG"] = function(args)
		TargetLog_Command(args);
	end
	SLASH_GRUDGE1="/grudge";
	SLASH_GRUDGE2="/kos";
	SlashCmdList["GRUDGE"] = function(args)
		TargetLog_Grudge(args);
	end
	-- Done!
	if (DEFAULT_CHAT_FRAME) then
		Loaded = "TargetLog v"..TLOG_VER_MAJOR.."."..TLOG_VER_MINOR.."."..TLOG_VER_REVIS.." with GrudgeList loaded";
		DEFAULT_CHAT_FRAME:AddMessage(Loaded);
	end
	UIErrorsFrame:AddMessage(Loaded, 1.0, 1.0, 1.0, 1.0, UIERRORS_HOLD_TIME);

end

function TargetLog_DataLoaded()
	TargetLog_VerifyData();

	-- Set up the tlog window
	TargetLogHistory_SetSize(TargetLogState.HistoryLines);
	TargetLogHistory_Update();
	if (TargetLogState.Visible == "true") then
		TargetLogHistoryFrame:Show();
		TargetLogHistoryFrameBackdropFrame:SetAlpha(TargetLogState.HistoryAlpha);
	else
		TargetLogHistoryFrameBackdropFrame:SetAlpha(TargetLogState.HistoryAlpha);
		TargetLogHistoryFrame:Hide();
	end
	TargetLog_ResetGuildFade();
end

function TargetLog_VerifyData()
	local key, value;
	local i;

	-- TargetLog data
	if not TargetLogData then
		TargetLog_ResetData();
	end
	for i = 1, TLOG_LOG_SIZE, 1 do
		if not TargetLogData[i] then TargetLogData[i] = { }; end
		if not TargetLogData[i].Level then TargetLogData[i].Level = TLOG_NO_LEVEL; end
		if not TargetLogData[i].Affiliation then TargetLogData[i].Affiliation = ""; end
		if not TargetLogData[i].Name then TargetLogData[i].Name = ""; end
		if not TargetLogData[i].Type then TargetLogData[i].Type = ""; end
		-- 0.9 - add guild info
		if not TargetLogData[i].Guild then TargetLogData[i].Guild = ""; end
		if not TargetLogData[i].RankName then TargetLogData[i].RankName = ""; end
		if not TargetLogData[i].RankNumber then TargetLogData[i].RankNumber = 0; end
		-- Remove Facing ... it'll be gone in the next patch.
		TargetLogData[i].Facing = nil;
	end

	-- Grudge list
	-- New in 0.6.0 - separate list per realm.

	if not TargetLogGrudge then
		TargetLogGrudge = { };
		TargetLogGrudge[RealmName] = { };
	else
		TargetLog_DataBackwardsCompat();
		if not TargetLogGrudge[RealmName] then
			-- Haven't made a list on this realm yet
			TargetLogGrudge[RealmName] = { };
		else
			-- Verify this realm's data
			TotalGrudges = 0;
			for key,value in TargetLogGrudge[RealmName] do
				if not TargetLogGrudge[RealmName][key].Name then
					TargetLogGrudge[RealmName][key] = nil;
				else
					if not TargetLogGrudge[RealmName][key].LastLevel then
						TargetLogGrudge[RealmName][key].LastLevel = "Unknown Level";
					end
					if not TargetLogGrudge[RealmName][key].Race then
						TargetLogGrudge[RealmName][key].Race = "Unknown Class";
					end
					if not TargetLogGrudge[RealmName][key].Class then
						TargetLogGrudge[RealmName][key].Class = "Unknown Class";
					end
					if not TargetLogGrudge[RealmName][key].Affiliation then
						TargetLogGrudge[RealmName][key].Affiliation = UNKNOWN_AFF;
					end
					if not TargetLogGrudge[RealmName][key].Reason then
						TargetLogGrudge[RealmName][key].Reason = "";
					end
					if not TargetLogGrudge[RealmName][key].LastSeen then
						TargetLogGrudge[RealmName][key].LastSeen = "Unknown";
					end
					if not TargetLogGrudge[RealmName][key].RankName then
						TargetLogGrudge[RealmName][key].RankName = "";
					end
					if not TargetLogGrudge[RealmName][key].RankNumber then
						TargetLogGrudge[RealmName][key].RankNumber = 0;
					end
					TotalGrudges = TotalGrudges + 1;
				end
			end
			TargetLogGrudge_Sort(CurrentSortType);
		end
	end

	-- Saved settings
	if not TargetLogState then TargetLogState = { }; end
	if not TargetLogState.Visible then TargetLogState.Visible = "true"; end
	if not TargetLogState.HistoryLines then TargetLogState.HistoryLines = TLOG_MAX_LOG_SHOW; end
	if not TargetLogState.HistoryAlpha then TargetLogState.HistoryAlpha = TLOG_DEFAULT_ALPHA; end
	if not TargetLogState.OnlyPC then TargetLogState.OnlyPC = "false"; end
	if not TargetLogState.OnlyPC then TargetLogState.OnlyHostile = "false"; end
	if not TargetLogState.Locked then TargetLogState.Locked = "false"; end
	if not TargetLogState.GuildShare then TargetLogState.GuildShare = "false"; end
	if not TargetLogState.GuildFade then TargetLogState.GuildFade = "false"; end
	if not TargetLogState.WarningChannel then TargetLogState.WarningChannel = "PARTY"; end
end

function TargetLog_DataBackwardsCompat()
	local key, value;
	local numnames;

	if ((not TargetLogGrudge) or TargetLogGrudge == { }) then return; end

	-- Pre-0.6.0, all data was indexed by TargetLogGrudge[key]
	-- Post-0.6.0, all data will be indexed by TargetLogGrudge[RealmName][key]
	numnames = 0;
	for key,value in TargetLogGrudge do
		if (TargetLogGrudge[key].Name) then numnames = numnames + 1; end
	end
	if (numnames > 0) then
		if (DEFAULT_CHAT_FRAME) then
			DEFAULT_CHAT_FRAME:AddMessage("Target Log: Updating Grudge List data from pre-0.6 to post-0.6 dataset");
		end
		-- Transfer TargetLogGrudge[key] over to TargetLogGrudge[RealmName][key]
		TargetLogGrudge[RealmName] = { };
		for key,value in TargetLogGrudge do
			if (key ~= RealmName) then
				TargetLogGrudge[RealmName][key] = { };
				TargetLogGrudge[RealmName][key].Name = TargetLogGrudge[key].Name;
				TargetLogGrudge[RealmName][key].LastLevel = TargetLogGrudge[key].LastLevel;
				TargetLogGrudge[RealmName][key].Race = TargetLogGrudge[key].Race;
				TargetLogGrudge[RealmName][key].Class = TargetLogGrudge[key].Class;
				TargetLogGrudge[RealmName][key].Affiliation = TargetLogGrudge[key].Affiliation;
				if TargetLogGrudge[key].LastGuild then
					TargetLogGrudge[RealmName][key].LastGuild = TargetLogGrudge[key].LastGuild;
				end
				TargetLogGrudge[key] = nil;
			end
		end
	end
end

function TargetLog_ResetData()
	local i;

	TargetLogData = { };
	for i = 1, TLOG_LOG_SIZE, 1 do
		TargetLogData[i] = { };
		TargetLogData[i].Level = TLOG_NO_LEVEL;
		--TargetLogData[i].Elite = "";
		TargetLogData[i].Affiliation = "";
		TargetLogData[i].Name = "";
		TargetLogData[i].Type = "";
		TargetLogData[i].Class = "";		-- Class or other?
		--TargetLogData[i].PvP = "";
		--TargetLogData[i].Warning = "";
		--TargetLogData[i].Facing = 0;
		TargetLogData[i].Guild = "";
		TargetLogData[i].LastSeen = "";
		--Added in 0.9.6
		TargetLogData[i].RankName = "";
		TargetLogData[i].RankNumber = 0;
	end
	CurrentGrudge = "";
end

function TargetLog_GetNameColour(Affiliation)
	local NameColour = DEFCOLOUR;

	if (Affiliation == FRIENDLY_GUILD) then NameColour = GUILDCOLOUR;
	elseif (Affiliation == FRIENDLY_PARTY) then NameColour = PARTYCOLOUR;
	elseif (Affiliation == NEUTRAL_NPC) then NameColour = NEUTRALCOLOUR;
	elseif (Affiliation == FRIENDLY_PC) then NameColour = FRIEND_PC_COLOUR;
	elseif (Affiliation == FRIENDLY_NPC) then NameColour = FRIEND_NPC_COLOUR;
	elseif (Affiliation == ENEMY_PC) then NameColour = ENEMY_PC_COLOUR;
	elseif (Affiliation == ENEMY_NPC) then NameColour = ENEMY_NPC_COLOUR;
	end
	return NameColour;
end

function TargetLog_UpdateList(arg1)
	local Level, Affiliation, Name, Type, Class, PvP, Guild, Elite;
	local LastSeen, RankName, RankNumber;
	local i, Bump;
	local checktype;

	if (TemporaryDisable) then
		return;
	end

	-- Checking target or mouseover?
	if (arg1 == "mouseover") then
		checktype = "mouseover";
	elseif (arg1 == "target") then
		checktype = "target";
	else
		return;
	end

	if not UnitName(checktype) then
		-- Make sure it's not a nonexistant unit
		return;
	end

	-- Check exceptions
	---- I see dead people...
	if (UnitIsDead(checktype)) then
		return;
	end
	---- We ignoring friendlies?
	if ((TargetLogState.OnlyHostile == "true") and (not UnitIsEnemy(checktype,"player"))) then
		-- Quick check for friendly grudges, even if we're not targetlogging them
		TargetLog_CheckAgainstGrudge(checktype);
		return;
	end
	---- We ignoring NPC's?
	if ((TargetLogState.OnlyPC == "true") and (not UnitIsPlayer(checktype))) then
		return;
	end

	-- Get Current Mouseover info
	Name = UnitName(checktype);
	Level = UnitLevel(checktype);

	-- Is it the player?
	if (Name == UnitName("player")) then
		return;
	end

	-- Friendliness
	if (UnitIsFriend(checktype,"player")) then
		if (UnitIsPlayer(checktype)) then
			Guild = GetGuildInfo(checktype);
			if (not Guild) then
				Guild = "";
			end
			if (UnitInParty(checktype)) then
				Affiliation = FRIENDLY_PARTY;
			else
				if (Guild and (Guild == GetGuildInfo("player"))) then
					Affiliation = FRIENDLY_GUILD;
				else
					Affiliation = FRIENDLY_PC;
				end
			end
		else
			Affiliation = FRIENDLY_NPC;
			Guild = "";
		end
	elseif (UnitIsEnemy(checktype,"player")) then
		if (UnitIsPlayer(checktype)) then
			Affiliation = ENEMY_PC;
			Guild = GetGuildInfo(checktype);
			if (not Guild) then
				Guild = "";
			end
		else
			Affiliation = ENEMY_NPC;
			Guild = "";
		end
	else
		Affiliation = NEUTRAL_NPC;
		Guild = "";
	end
	-- 'Type' (ie. Race/Humanoid/Beast/Dragonkin/etc.)
	if (UnitIsPlayer(checktype)) then
		Type = UnitRace(checktype);	
	else
		Type = UnitCreatureType(checktype);
		--Type = string.gsub(Type,"Humanoid","H'noid");
		--Type = string.gsub(Type,"Mechanical","Mech.");
		--Type = string.gsub(Type,"Elemental","Elem.");
		if Type then
			Type = string.gsub(Type,"Not specified","");
		else
			Type = "";
		end
	end
	-- Class
	if (UnitIsPlayer(checktype)) then
		Class = UnitClass(checktype);
		Elite = nil;
	else
		Class = "";
		--if (UnitIsPlusMob(checktype) or	-- Removed function as of interface 1300
		if (UnitClassification(checktype) == "elite" or
		    UnitClassification(checktype) == "worldboss") then
			Elite = "true";
		else
			Elite = nil;
		end
	end
	-- PvP status
	if (UnitIsPVP(checktype)) then
		PvP = true;
	else
		PvP = nil;
	end

	-- LastSeen
	LastSeen = date();

	-- Rank
	if UnitIsPlayer(checktype) and UnitPVPRank(checktype) then
		RankName, RankNumber = GetPVPRankInfo(UnitPVPRank(checktype), checktype);
	else
		RankName = "";
		RankNumber = 0;
	end

	-- Facing direction
	--Facing = math.deg(UnitFacing(checktype));	-- Removed function as of interface 1300

	-- Add it to the list if it is there, otherwise, update list entry
	i = 1;
	Bump = TLOG_LOG_SIZE;
	while (i <= TLOG_LOG_SIZE) do
		if ((TargetLogData[i].Name == Name) and
		    (TargetLogData[i].Level == Level) and 
		    (TargetLogData[i].Type == Type)) then
			-- This one matches current entry.
			Bump = i;
			i = TLOG_LOG_SIZE + 1;
		end
		i = i + 1;
	end
	-- So everything from Bump upwards gets shifted to overwrite Bump
	-- If mouseover did not match, Bump = last entry
	i = Bump;
	while (i >= 2) do
		TargetLogData[i].Level = TargetLogData[i-1].Level;
		TargetLogData[i].Elite = TargetLogData[i-1].Elite;
		TargetLogData[i].Affiliation = TargetLogData[i-1].Affiliation;
		TargetLogData[i].Name = TargetLogData[i-1].Name;
		TargetLogData[i].Type = TargetLogData[i-1].Type;
		TargetLogData[i].Class = TargetLogData[i-1].Class;
		TargetLogData[i].PvP = TargetLogData[i-1].PvP;
		--TargetLogData[i].Facing = TargetLogData[i-1].Facing;
		TargetLogData[i].Warning = TargetLogData[i-1].Warning;
		TargetLogData[i].Guild = TargetLogData[i-1].Guild;
		TargetLogData[i].LastSeen = TargetLogData[i-1].LastSeen;
		TargetLogData[i].RankName = TargetLogData[i-1].RankName;
		TargetLogData[i].RankNumber = TargetLogData[i-1].RankNumber;
		i = i - 1;
	end
	-- Put the new one at the top.
	TargetLogData[1].Level = Level;
	TargetLogData[1].Elite = Elite;
	TargetLogData[1].Affiliation = Affiliation;
	TargetLogData[1].Name = Name;
	TargetLogData[1].Type = Type;
	TargetLogData[1].Class = Class;
	TargetLogData[1].PvP = PvP;
	--TargetLogData[1].Facing = Facing;
	TargetLogData[1].Guild = Guild;
	TargetLogData[1].LastSeen = LastSeen;
	TargetLogData[1].RankName = RankName;
	TargetLogData[1].RankNumber = RankNumber;

	if (UnitIsPlayer(checktype)) then
		-- Check grudge list, now that we have name/level/class/race in slot 1, so we don't waste time recollecting
		if (TargetLog_CheckAgainstGrudge(checktype) == "true") then
			TargetLogData[1].Warning = "true";
		else
			TargetLogData[1].Warning = nil;
		end
	else
		TargetLogData[1].Warning = nil;
	end
end

function TargetLog_GenerateChatString(number)
	local flags = ""; 
	local level = "";
	local chatstr;
	
	-- Elite check
	if (TargetLogData[number].Elite) then
		level = TargetLogData[number].Level.." Elite";
	else
		level = TargetLogData[number].Level;
	end

	-- Other flags
	--if (not TargetLogData[number].PvP) then
	--	flags = ", PvP enabled";
	--end

	-- Generate the string
	chatstr = string.format("%s, L%s %s %s",
	                        TargetLogData[number].Name,
	                        level,
	                        TargetLogData[number].Type,
	                        TargetLogData[number].Class);
	--                        TargetLog_DirectionString(TargetLogData[number].Facing));
	chatstr = chatstr..flags.." ";
	return chatstr;
end

function TargetLog_Command(args)
	local opts, i;
	local msgs = {"\n"};
	local alpha;
	local erred = nil;
	local tlogheader = "|cffff1f1fTargetLog v"..TLOG_VER_MAJOR.."."..TLOG_VER_MINOR.."."..TLOG_VER_REVIS;

	if (args) then
		opts = string.lower(args);
		i = tonumber(args);
		if (opts == "") then
			--DEFAULT_CHAT_FRAME:AddMessage("|cffffff00TargetLog v"..VER_MAJOR.."."..VER_MINOR.."."..VER_REVIS);
			table.insert(msgs,tlogheader);
			table.insert(msgs,"Type: |cff00ff00/tlog help|r for list of commands.");
		elseif (opts == "help") then
			table.insert(msgs,tlogheader);
			table.insert(msgs,"Usage: |cff00ff00/tlog <option>");
			table.insert(msgs,"Options:");
			table.insert(msgs,"     |cff00ff00clear|r - clear log");
			table.insert(msgs,"     |cff00ff00onlypc|r/|cff00ff00notonlypc|r - show only PC's/all");
			table.insert(msgs,"     |cff00ff00onlyhostile|r/|cff00ff00notonlyhostile|r - show only hostiles/all");
			table.insert(msgs,"     |cff00ff00onlyhostile|r/|cff00ff00notonlyhostile|r - show only hostiles/all");
			table.insert(msgs,"     |cff00ff00guildfade|r/|cff00ff00noguildfade|r - enable/disable fade between player class and guild in history window");
			table.insert(msgs,"     |cff00ff00grudgeshare|r/|cff00ff00nogrudgeshare|r - share grudge additions with guild or not");
			table.insert(msgs,"     |cff00ff00<numlines>|r - set lines in TargetLog history window (1-16)");
			table.insert(msgs,"     |cff00ff00alpha <value>|r - set transparency of TargetLog history window (0.0-1.0)");
			table.insert(msgs,"     |cff00ff00channel (party/raid/guild/localdefense)|r - set warning channel");
			table.insert(msgs,"|cff00ff00/kos <name>|r to add player, or use |cff00ff00!<name>|r to remove player");
			erred = 1;
		elseif (opts == "clear") then
			TargetLog_ResetData();
			TargetLogHistory_Display();
			table.insert(msgs,"TargetLog: Cleared");
			erred = 1;
		elseif (opts == "onlypc") then
			TargetLogState.OnlyPC = "true";
			TargetLogToggleOnlyPC:SetChecked(1);
		elseif (opts == "notonlypc") then
			TargetLogState.OnlyPC = "false";
			TargetLogToggleOnlyPC:SetChecked(0);
		elseif (opts == "onlyhostile") then
			TargetLogState.OnlyHostile = "true";
			TargetLogToggleOnlyHostile:SetChecked(1);
		elseif (opts == "notonlyhostile") then
			TargetLogState.OnlyHostile = "false";
			TargetLogToggleOnlyHostile:SetChecked(0);
		elseif (opts == "lock") then
			TargetLogState.Locked = "true";
			TargetLogHistory_OnDragStop();
		elseif (opts == "unlock") then
			TargetLogState.Locked = "false";
		elseif (opts == "guildfade") then
			TargetLogState.GuildFade = "true";
			TargetLog_ResetGuildFade();
		elseif (opts == "noguildfade") then
			TargetLogState.GuildFade = "false";
			TargetLog_ResetGuildFade();
		elseif (opts == "guildshare") or (opts == "grudgeshare") then
			TargetLogState.GuildShare = "true";
			TargetLogGrudgeGuildShareToggle:SetChecked(1);
		elseif (opts == "noguildshare") or (opts == "nogrudgeshare") then
			TargetLogState.GuildShare = "false";
			TargetLogGrudgeGuildShareToggle:SetChecked(0);
		elseif (opts == "nosound") then
			TargetLogState.NoSound = "true";
		elseif (opts == "sound") then
			TargetLogState.NoSound = nil;
		elseif (string.find(opts,"alpha ") == 1) then
			alpha = tonumber(string.sub(opts,7,string.len(opts)));
			if (alpha and ((alpha >= 0.0) and (alpha <= 1.0))) then
				TargetLogState.HistoryAlpha = alpha;
				TargetLogHistoryFrameBackdropFrame:SetAlpha(alpha);
			else
				table.insert(msgs,string.format("|cffff0000Invalid alpha specified. Valid range: 0.0 to 1.0",alpha));
				erred = 1;
			end
		elseif (string.find(opts,"channel ") == 1) then
			local channelstring = string.lower(string.sub(opts,9,string.len(opts)));
			if (channelstring == "party") then
				TargetLogState.WarningChannel = "PARTY";
			elseif (channelstring == "raid") then
				TargetLogState.WarningChannel = "RAID";
			elseif (channelstring == "guild") then
				TargetLogState.WarningChannel = "GUILD";
			elseif (channelstring == "localdefense") then
				TargetLogState.WarningChannel = "LocalDefense";
			elseif (channelstring == "smart") then
				TargetLogState.WarningChannel = "SMART";
			else
				table.insert(msgs,string.format("|cffff0000Invalid channel specified. Valid channels are (party/raid/guild/localdefense/smart)",alpha));
				erred = 1;
			end
		elseif (i) then
			if ((i > 0) and (i <= TLOG_MAX_LOG_SHOW)) then
				TargetLogState.HistoryLines = i;
				TargetLogHistory_SetSize(i);
			else
				table.insert(msgs,"|cffff0000Invalid number of lines specified: "..args..". Valid range: 1 to "..TLOG_MAX_LOG_SHOW);
				erred = 1;
			end
		else
			table.insert(msgs,tlogheader);
			table.insert(msgs,"Unknown option: |cff00ff00"..args);
			table.insert(msgs,"Type: |cff00ff00/tlog help|r for list of commands.");
			erred = 1;
		end
		for i=1,table.getn(msgs),1 do
			DEFAULT_CHAT_FRAME:AddMessage(msgs[i]);
		end
		if not erred then TargetLog_ShowState(); end
	end
end

function TargetLog_ToggleHandle(arg1)
	local BState;
	if (arg1) then
		if (arg1 == "OnlyPC") then
			BState = TargetLogToggleOnlyPC:GetChecked();
			if (BState) then
				TargetLogState.OnlyPC = "true";
			else
				TargetLogState.OnlyPC = "false";
			end
		elseif (arg1 == "OnlyHostile") then
			BState = TargetLogToggleOnlyHostile:GetChecked();
			if (BState) then
				TargetLogState.OnlyHostile = "true";
			else
				TargetLogState.OnlyHostile = "false";
			end
		elseif (arg1 == "GuildShare") then
			BState = TargetLogGrudgeGuildShareToggle:GetChecked();
			if (BState) then
				TargetLogState.GuildShare = "true";
			else
				TargetLogState.GuildShare = "false";
			end
		end
	end
end

function TargetLog_ShowState()
	local msgs = {"TargetLog/GrudgeList current settings:"};
	local str = "";

	if (TargetLogState.Locked == "true") then
		table.insert(msgs,"     |cffffff00Log window is locked.");
	else
		table.insert(msgs,"     |cffffff00Log window is unlocked.");
	end

	table.insert(msgs,"     |cffffff00TargetLog window size: "..TargetLogState.HistoryLines);
	table.insert(msgs,string.format("     |cffffff00TargetLog background alpha: %.2f",TargetLogState.HistoryAlpha));

	if (TargetLogState.OnlyPC == "true") then
		table.insert(msgs,"     |cffffff00Logging PC's only.");
	else
		table.insert(msgs,"     |cffffff00Logging PC's and NPC's.");
	end
	if (TargetLogState.OnlyHostile == "true") then
		table.insert(msgs,"     |cffffff00Logging hostiles only.");
	else
		table.insert(msgs,"     |cffffff00Logging hostiles and non-hostiles.");
	end
	if (TargetLogState.GuildFade == "true") then
		table.insert(msgs,"     |cffffff00Fading between race/class and guild in history window.");
	else
		table.insert(msgs,"     |cffffff00History window shows race/class only.");
	end
	if (TargetLogState.GuildShare == "true") then
		table.insert(msgs,"     |cffffff00Guild sharing of GrudgeList adds is enabled.");
	else
		table.insert(msgs,"     |cffffff00Guild sharing of GrudgeList adds is disabled.");
	end
	if (TargetLogState.NoSound and TargetLogState.NoSound == "true") then
		table.insert(msgs,"     |cffffff00Audible warning disabled.");
	else
		table.insert(msgs,"     |cffffff00Audible warning enabled.");
	end

	table.insert(msgs,"     |cffffff00Current warning channel is: "..TargetLogState.WarningChannel);
	
	-- Output it
	for i=1,table.getn(msgs),1 do
		DEFAULT_CHAT_FRAME:AddMessage(msgs[i]);
	end
end

-- This function is no longer useful as of 1.3.0

--function TargetLog_DirectionString(facing)
--	local dirstring = "";
--	if (((facing > 337.5) and (facing <= 360)) or
--	    ((facing > 0) and (facing <= 23.5))) then
--		dirstring = "N";
--	elseif ((facing > 23.5) and (facing <= 68.5)) then
--		dirstring = "NW";
--	elseif ((facing > 68.5) and (facing <= 113.5)) then
--		dirstring = "W";
--	elseif ((facing > 113.5) and (facing <= 158.5)) then
--		dirstring = "SW";
--	elseif ((facing > 158.5) and (facing <= 203.5)) then
--		dirstring = "S";
--	elseif ((facing > 203.5) and (facing <= 248.5)) then
--		dirstring = "SE";
--	elseif ((facing > 248.5) and (facing <= 293.5)) then
--		dirstring = "E";
--	elseif ((facing > 293.5) and (facing <= 337.5)) then
--		dirstring = "NE";
--	end
--	return dirstring;
--end

function TargetLog_OnUpdate(elapsed)
	if (not TargetLogHistoryFrame:IsVisible()) then
		return;
	end

	LastUpdateTime = LastUpdateTime - elapsed;
	if (LastUpdateTime <= 0) then
		LastUpdateTime = TLOG_ONUPDATE_TIME;
		-- Workaround for non-mouseover target changes not triggering PLAYER_TARGET_CHANGED event
		if (LastTargetName ~= UnitName("target")) then
			LastTargetName = UnitName("target");
			TargetLog_UpdateList("target");
			TargetLogHistory_Update();
			TargetLogGrudge_UpdateButtonStates()
		end
	end

	-- Field fading
	if (TargetLogState.GuildFade == "true") then
		local i;
		LastFieldUpdateTime = LastFieldUpdateTime - elapsed;
		if (LastFieldUpdateTime <= 0) then
			LastFieldUpdateTime = TLOG_FIELDUPDATE_TIME
			if (FieldDisplay == "class") then
				-- Fade out desc, fade in desc2
				for i=1,TLOG_MAX_LOG_SHOW,1 do
					UIFrameFadeOut(getglobal("TargetLogHistoryButton"..i.."Description"), TLOG_FIELDFADETIME, 1.0, 0.0);
					UIFrameFadeIn(getglobal("TargetLogHistoryButton"..i.."Description2"), TLOG_FIELDFADETIME, 0.0, 1.0);
				end
				FieldDisplay = "guild";
			else
				-- Fade out desc2, fade in desc1
				for i=1,TLOG_MAX_LOG_SHOW,1 do
					UIFrameFadeOut(getglobal("TargetLogHistoryButton"..i.."Description2"), TLOG_FIELDFADETIME, 1.0, 0.0);
					UIFrameFadeIn(getglobal("TargetLogHistoryButton"..i.."Description"), TLOG_FIELDFADETIME, 0.0, 1.0);
				end
				FieldDisplay = "class";
			end
		end
	end

end

function TargetLog_ResetGuildFade()
	local i;

	for i = 1,TLOG_MAX_LOG_SHOW,1 do	
		UIFrameFadeOut(getglobal("TargetLogHistoryButton"..i.."Description2"), TLOG_FIELDFADETIME, 0.0, 0.0);
		UIFrameFadeIn(getglobal("TargetLogHistoryButton"..i.."Description"), TLOG_FIELDFADETIME, 0.0, 1.0);
		FieldDisplay = "class";
	end
end

--==========================================================================--
-- Functions specific to the History window

function TargetLogHistory_SetSize(lines)
	local i;
	local button;
	local height;

	if ((lines < 1) or (lines > TLOG_MAX_LOG_SHOW)) then
		return;
	end

	for i=1,TLOG_MAX_LOG_SHOW,1 do
		button=getglobal("TargetLogHistoryButton"..i);
		if (i <= TargetLogState.HistoryLines) then
			button:Show();
		else
			button:Hide();
		end
	end
	height =	(TargetLogHistoryButton1:GetHeight() * TargetLogState.HistoryLines)
			+ TargetLogHistoryFrameHeaderFrame:GetHeight()
			+ 6;
	TargetLogHistoryFrame:SetHeight(height);
	TargetLogHistoryFrameBackdropFrame:SetHeight(height);
	--TargetLogHistoryFrameBackdropFrame:SetHeight((TargetLogHistoryButton1:GetHeight() * TargetLogState.HistoryLines) + 6);
	TargetLogHistory_Update();
	--TargetLog_ResetData();
end

function TargetLogHistory_Toggle()
	-- Toggle visibility
	if (TargetLogState.Visible == "true") then
		TargetLogHistoryFrame:Hide();
		TargetLogState.Visible = "false";
		PlaySound("igQuestLogClose");
	else 
		TargetLogHistoryFrame:Show();
		TargetLogState.Visible = "true";
		PlaySound("igQuestLogOpen");
	end
end

function TargetLogHistory_OnDragStart()
	if(TargetLogState.Locked == "false") then
		TargetLogHistoryFrame:StartMoving();
		--TargetLogHeader:SetText("Moving");
        end
end

function TargetLogHistory_OnDragStop()
        TargetLogHistoryFrame:StopMovingOrSizing();
	--TargetLogHeader:SetText("TargetLog");
end

function TargetLogHistory_Update()
	local Text = "";
	local Diff, Name;
	local TName, TLevel;
	local DelimL, DelimR;
	local Type;
	
	local button;
	local i, diff;

	local numstring = "";
	local numcolour = "|cffffff00";
	local tagstring = "";
	local dirstring = "";
	local namecolour = DEFCOLOUR;
	local desccolour = DEFCOLOUR;
	
	if UnitName("target") then
		TName = UnitName("target");
		TLevel = UnitLevel("target");
	else
		TName = ""; TLevel = "";
	end

	-- Update the display frame
	for i=1,TargetLogState.HistoryLines,1 do
		button = getglobal("TargetLogHistoryButton"..i);
		if (TargetLogData[i].Level == TLOG_NO_LEVEL) then
			button:Hide();
		else
			button:Show();

			-- Level number
			if (TargetLogData[i].Level ~= -1) then
				diff = UnitLevel("Player") - TargetLogData[i].Level;
				-- if (diff > 7) then
				if (diff > 12) then
					numcolour = "|cffafafaf";
				-- elseif (diff >= 3) then 
				elseif (diff >= 6) then 
					numcolour = "|cff00ff00";
				-- elseif (diff >= -2) then
				--	numcolour = "|cffffff00";
				-- elseif (diff >= -4) then
				--	numcolour = "|cffff7f00";
				elseif (diff < 6) then
					numcolour = "|cffefbf00";
				else
					-- Too high
					numcolour = "|cffff0000";
				end
				numstring = numcolour..TargetLogData[i].Level;
			else
				-- -1 level, ie. too high, so ??
				numstring = "|cffff0000??";
			end
			getglobal("TargetLogHistoryButton"..i.."Level"):SetText(numstring);

			-- Tags (PvP, Elite, etc).
			tagstring = "";
			if (TargetLogData[i].Elite) then -- Elite
				tagstring = tagstring.."|cffffff1f+";
			end
			if (TargetLogData[i].PvP) then -- PvP
				tagstring = tagstring.."|cffff4f4fP";
			end
			getglobal("TargetLogHistoryButton"..i.."Tag"):SetText(tagstring);

			-- Name
			namecolour = TargetLog_GetNameColour(TargetLogData[i].Affiliation);
			getglobal("TargetLogHistoryButton"..i.."Name"):SetText(namecolour..TargetLogData[i].Name);

			-- Type & Class
			if ((TargetLogData[i].Affiliation == FRIENDLY_PC) or
			    (TargetLogData[i].Affiliation == ENEMY_PC) or 
			    (TargetLogData[i].Affiliation == FRIENDLY_PARTY) or 
			    (TargetLogData[i].Affiliation == FRIENDLY_GUILD)) then
				desccolour = HICOLOUR;
			else
				desccolour = DEFCOLOUR;
			end

			Type = TargetLogData[i].Type;
			Type = string.gsub(Type,"^Human$","Hum.");
			Type = string.gsub(Type,"^Night Elf$","N.Elf");
			Type = string.gsub(Type,"^Gnome$","Gnom.");
			--Type = string.gsub(Type,"Dwarf","Dwarf");
			--Orc should be okay
			Type = string.gsub(Type,"^Tauren$","Taur.");
			Type = string.gsub(Type,"^Undead$","Und.");

			getglobal("TargetLogHistoryButton"..i.."Description"):SetText(desccolour..Type.." "..TargetLogData[i].Class);

			-- Description2 field - if non-player or no guild, set same
			if ((TargetLogData[i].Guild) and (TargetLogData[i].Guild ~= "")) then
				getglobal("TargetLogHistoryButton"..i.."Description2"):SetText("|cffffaf00<"..TargetLogData[i].Guild..">");
			else
				getglobal("TargetLogHistoryButton"..i.."Description2"):SetText(desccolour..Type.." "..TargetLogData[i].Class);
			end

			-- Flags
			-- Facing -- Disabled for now ... only useful for those with a very good sense of spatial orientation
			--dirstring = TargetLog_DirectionString(TargetLogData[i].Facing);
			--getglobal("TargetLogHistoryButton"..i.."Flags"):SetText("|cff1f7fff"..dirstring);

			getglobal("TargetLogHistoryButton"..i.."Flags"):SetText("");

			-- Highlight if warning
			if (TargetLogData[i].Warning) then
				getglobal("TargetLogHistoryButton"..i.."Highlight"):SetVertexColor(VERT_WARN_R, VERT_WARN_G, VERT_WARN_B);
				button:LockHighlight();
			elseif ((TargetLogData[i].Name == TName) and (TargetLogData[i].Level == TLevel)) then
				getglobal("TargetLogHistoryButton"..i.."Highlight"):SetVertexColor(VERT_TARG_R, VERT_TARG_G, VERT_TARG_B);
				button:LockHighlight();
			else
				getglobal("TargetLogHistoryButton"..i.."Highlight"):SetVertexColor(VERT_NORM_R, VERT_NORM_G, VERT_NORM_B);
				button:UnlockHighlight();
			end

		end
	end

	if (TargetLogState.OnlyPC == "true") then
		TargetLogToggleOnlyPC:SetChecked(1);
	else
		TargetLogToggleOnlyPC:SetChecked(0);
	end
	if (TargetLogState.OnlyHostile == "true") then
		TargetLogToggleOnlyHostile:SetChecked(1);
	else
		TargetLogToggleOnlyHostile:SetChecked(0);
	end
end

function TargetLogHistory_ListButton()
	local name;
	local chatstr;

	local cltime = GetTime();
	local i = this:GetID();
	local wantname = TargetLogData[i].Name;
	
	-- Handle double-click
	if ((DoubleClick[1] == id) and ((cltime - DoubleClick[2]) < TLOG_DOUBLECLICK_TIME)) then
		chatstr = TargetLog_GenerateChatString(i);
		if (TargetLogState.WarningChannel == "LocalDefense") then
			-- Borrowed from WatchTower
			local temparray = { GetMapZones(GetCurrentMapContinent()) };
			local mapzone = temparray[GetCurrentMapZone()];
			local localdef = GetChannelName("LocalDefense - "..mapzone);
			SendChatMessage(chatstr,"CHANNEL","Common",localdef);
		elseif (TargetLogState.WarningChannel == "SMART") then
			if (GetNumRaidMembers() > 0) then
				SendChatMessage(chatstr, "RAID");
			elseif (GetNumPartyMembers() > 0) then
				SendChatMessage(chatstr, "PARTY");
			end
		else
			SendChatMessage(chatstr,TargetLogState.WarningChannel);
		end
	else
		-- Single click, try to target;
		if (IsAltKeyDown() and IsControlKeyDown()) then
			-- Add the entry to the grudge list.
			TargetLog_AddGrudge(wantname, "false");
		elseif IsShiftKeyDown() then
			if ChatFrameEditBox:IsVisible() then
				chatstr = TargetLog_GenerateChatString(i)
				ChatFrameEditBox:Insert(chatstr);
			else
				-- Ignore it, so that we can double-click without it scrolling from target change
			end
		else
			--DEFAULT_CHAT_FRAME:AddMessage("Trying to target "..wantname);
			TargetByName(wantname);
	
			name = UnitName("target")
			if (name ~= wantname) then
				ClearTarget();
				--TargetLastEnemy();
			end
			--TargetLogGrudge_UpdateButtonStates()
			TargetLog_UpdateList("target");
		end
		TargetLogHistory_Update();
	end
	DoubleClick = {id, cltime};
end

function TargetLogHistory_ButtonOnEnter()
	local i = this:GetID();
	local name = TargetLogData[i].Name;
	local key = string.lower(name);
	local level = TargetLogData[i].Level;
	local race = TargetLogData[i].Type;
	local class = TargetLogData[i].Class;
	local guild, rank, infoline;

	if (TargetLogData[i].Guild and TargetLogData[i].Guild ~= "") then
		guild = "<"..TargetLogData[i].Guild..">";
	else
		guild = "";
	end

	if (TargetLogData[i].RankNumber > 0) then
		rank = TargetLogData[i].RankName.."("..TargetLogData[i].RankNumber..")";
	end

	infoline = level.." "..race.." "..class;

	-- Option to add PvP data later 

	GameTooltip:SetOwner(getglobal("TargetLogHistoryButton"..i),"ANCHOR_TOPLEFT");
	--GameTooltip:SetText("Reason:",1.0,1.0,1.0,1);
	--GameTooltip:SetText(name..guild,1.0,1.0,1.0,1);
	GameTooltip:AddDoubleLine(name, infoline ,1.0,1.0,1.0, 1.0,0.75,0.0);
	if (guild ~= "") and (rank ~= "") then
		GameTooltip:AddDoubleLine(guild, rank, 1.0, 1.0, 0.25, 1.0, 1.0, 1.0);
	end

	if (TargetLogGrudge[RealmName][key] and TargetLogGrudge[RealmName][key].Reason) then
		GameTooltip:AddLine(TargetLogGrudge[RealmName][key].Reason,1.0,1.0,0.25,1);
	end
	GameTooltip:Show();
end

--==========================================================================--
-- Grudgecode

function TargetLog_Grudge(args)
	local opts, name;

	if (args) then
		opts = string.lower(args);
		if ((opts == "") or (opts == "list")) then
			TargetLog_ListGrudges();
		elseif (opts == "add") then
			-- Do we have a current target?
			if (UnitName("target") and (UnitName("target") ~= UnitName("player"))) then
				if (UnitIsPlayer("target")) then
					TargetLog_AddGrudge(UnitName("target"), "true");
				else
					DEFAULT_CHAT_FRAME:AddMessage("|cffff1f1f"..UnitName("target").." is not a player. Cannot add to grudge list");
				end
			else
				DEFAULT_CHAT_FRAME:AddMessage("|cffff1f1fNo target currently selected.");
			end
		else
			if (string.find(opts,"!") == 1) then
				-- Add or remove?
				name = string.sub(opts,2,string.len(opts))
				if (name ~= "") then
					TargetLog_RemoveGrudge(name,0);
				end
			else
				TargetLog_AddGrudge(opts, "false");
			end
		end
	end
end

function TargetLog_ListGrudges()
	local NumGrudge;
	local showstr = "";

	-- Big long grudge list.
	DEFAULT_CHAT_FRAME:AddMessage("\n|cffff1f1fGrudge list (Realm: "..RealmName.."):");
	NumGrudge = 0;
	for key,value in TargetLogGrudge[RealmName] do
		showstr = TargetLog_GetNameColour(TargetLogGrudge[RealmName][key].Affiliation)..TargetLogGrudge[RealmName][key].Name;
		showstr = showstr..GRUDGELISTCOLOUR..", "..TargetLogGrudge[RealmName][key].LastLevel;
		showstr = showstr.." "..TargetLogGrudge[RealmName][key].Race;
		showstr = showstr.." "..TargetLogGrudge[RealmName][key].Class;
		if TargetLogGrudge[RealmName][key].LastGuild then
			showstr = showstr.." <"..TargetLogGrudge[RealmName][key].LastGuild..">";
		end
		DEFAULT_CHAT_FRAME:AddMessage(showstr);
		NumGrudge = NumGrudge + 1;
	end
	TotalGrudges = NumGrudge;
	DEFAULT_CHAT_FRAME:AddMessage("|cffff1f1f"..NumGrudge.." player(s) on list.");
end

function TargetLog_AddGrudge(namestr, havetarget, lumpdata)
	-- IMPORTANT: All indexing names are lower case
	-- All PRINTED names are FIRSTER LETTER capitalized
	local showstr = "";
	local name = string.lower(namestr);
	local i, inlog;
	local sharestr = "";
	local RankNumber, RankName;

	if TargetLogGrudge[RealmName][name] then
		if (not lumpdata) then
			DEFAULT_CHAT_FRAME:AddMessage("|cffff1f1f"..TargetLogGrudge[RealmName][name].Name.." is already on the grudge list.");
		else
			DEFAULT_CHAT_FRAME:AddMessage("|cffff1f1f"..TargetLogGrudge[RealmName][name].Name.." is already on the grudge list. Updating information.");
			-- Update via transmitted data
			TargetLogGrudge[RealmName][name].LastLevel = lumpdata.LastLevel;
			TargetLogGrudge[RealmName][name].Race = lumpdata.Race;
			TargetLogGrudge[RealmName][name].Class = lumpdata.Class;
			TargetLogGrudge[RealmName][name].Affiliation = lumpdata.Affiliation;
			if (lumpdata.Guild) then
				TargetLogGrudge[RealmName][name].Guild = lumpdata.Guild;
			end
			-- We won't overwrite these two if someone is already on our list.
			-- TargetLogGrudge[RealmName][name].LastSeen = "Unknown";
			-- TargetLogGrudge[RealmName][name].Reason = lumpdata.Reason;
		end
	else
		TargetLogGrudge[RealmName][name] = { };
		TargetLogGrudge[RealmName][name].Name = string.gsub(namestr,"^%l",string.upper);
		if (not lumpdata) then
			if (havetarget == "false") then
				-- We can check through TargetLog data for level, race and class if it's in there
				i = 1; inlog = 0;
				while ((i <= TLOG_LOG_SIZE) and (inlog == 0)) do
					if (name == string.lower(TargetLogData[i].Name)) then
						inlog = i;
					end
					i = i + 1;
				end
				if (inlog > 0) then
					if ((TargetLogData[inlog].Affiliation == ENEMY_PC) or 
					    (TargetLogData[inlog].Affiliation == FRIENDLY_PC)) then
						TargetLogGrudge[RealmName][name].Name = TargetLogData[inlog].Name;
						TargetLogGrudge[RealmName][name].Race = TargetLogData[inlog].Type;
						TargetLogGrudge[RealmName][name].Class = TargetLogData[inlog].Class;
						TargetLogGrudge[RealmName][name].LastLevel = TargetLogData[inlog].Level;
						TargetLogGrudge[RealmName][name].Affiliation = TargetLogData[inlog].Affiliation;
						-- We now have guild info... sometimes;
						if (TargetLogData[inlog].Guild and (TargetLogData[inlog].Guild ~= "")) then
							TargetLogGrudge[RealmName][name].LastGuild = TargetLogData[inlog].Guild;
						end
						TargetLogGrudge[RealmName][name].LastSeen = TargetLogData[inlog].LastSeen;
						TargetLogGrudge[RealmName][name].RankName = TargetLogData[inlog].RankName;
						TargetLogGrudge[RealmName][name].RankNumber = TargetLogData[inlog].RankNumber;
					else
						-- It's not a player!
						DEFAULT_CHAT_FRAME:AddMessage("|cffff1f1f"..TargetLogGrudge[RealmName][name].Name.." is not a player. Cannot add to grudge list");
						TargetLogGrudge[RealmName][name] = nil;
						return;
					end
				else
					-- TODO: Possibly we could use the who list to find?
					-- Maybe ... maybe not. Server how often you can get who info
					TargetLogGrudge[RealmName][name].Race = "Unknown race";
					TargetLogGrudge[RealmName][name].Class = "Unknown class";
					TargetLogGrudge[RealmName][name].LastLevel = "Unknown level";
					TargetLogGrudge[RealmName][name].Affiliation = UNKNOWN_AFF;
					TargetLogGrudge[RealmName][name].LastGuild = nil;
					TargetLogGrudge[RealmName][name].LastSeen = "Unknown";
					TargetLogGrudge[RealmName][name].RankName = "";
					TargetLogGrudge[RealmName][name].RankNumber = 0;
				end
			else
				TargetLogGrudge[RealmName][name].Name = UnitName("target");
				TargetLogGrudge[RealmName][name].Race = UnitRace("target");
				TargetLogGrudge[RealmName][name].Class = UnitClass("target");
				TargetLogGrudge[RealmName][name].LastLevel = UnitLevel("target");
				TargetLogGrudge[RealmName][name].LastGuild = GetGuildInfo("target");
				if (UnitIsEnemy("target","player")) then
					TargetLogGrudge[RealmName][name].Affiliation = ENEMY_PC;
				else
					TargetLogGrudge[RealmName][name].Affiliation = FRIENDLY_PC;
				end
				TargetLogGrudge[RealmName][name].LastSeen = date();
				if UnitPVPRank("target") then
					RankName, RankNumber = GetPVPRankInfo(UnitPVPRank,"target");
					TargetLogGrudge[RealmName][name].RankName = RankName;
					TargetLogGrudge[RealmName][name].RankNumber = RankNumber;
				else
					TargetLogGrudge[RealmName][name].RankName = "";
					TargetLogGrudge[RealmName][name].RankNumber = 0;
				end
			end
			TargetLogGrudge[RealmName][name].Reason = "";
		else
			-- Add via transmitted data
			TargetLogGrudge[RealmName][name].Name = lumpdata.Name;
			TargetLogGrudge[RealmName][name].LastLevel = lumpdata.LastLevel;
			TargetLogGrudge[RealmName][name].Race = lumpdata.Race;
			TargetLogGrudge[RealmName][name].Class = lumpdata.Class;
			TargetLogGrudge[RealmName][name].Affiliation = lumpdata.Affiliation;
			if (lumpdata.Guild) then
				TargetLogGrudge[RealmName][name].Guild= lumpdata.Guild;
			end
			TargetLogGrudge[RealmName][name].Reason = lumpdata.Reason;
			TargetLogGrudge[RealmName][name].LastSeen = "Unknown";
			TargetLogGrudge[RealmName][name].RankName = "";
			TargetLogGrudge[RealmName][name].RankNumber = 0;
		end

		-- Print a confirmation
		showstr = TargetLog_GetNameColour(TargetLogGrudge[RealmName][name].Affiliation)..TargetLogGrudge[RealmName][name].Name;
		showstr = showstr..GRUDGELISTCOLOUR..", "..TargetLogGrudge[RealmName][name].LastLevel;
		showstr = showstr.." "..TargetLogGrudge[RealmName][name].Race;
		showstr = showstr.." "..TargetLogGrudge[RealmName][name].Class;
		if TargetLogGrudge[RealmName][name].LastGuild then
			showstr = showstr.." <"..TargetLogGrudge[RealmName][name].LastGuild..">";
		end
		DEFAULT_CHAT_FRAME:AddMessage(showstr.." added to grudge list.");
		TotalGrudges = TotalGrudges - 1;

		-- Update UI
		TargetLogGrudgeFrame.ListSelected = -1;
		GrudgePanelSelect = nil;
		TargetLogGrudge_Sort(CurrentSortType);
		TargetLogGrudge_UpdateList();
		
		-- GuildShare do only if not received from guildshareBugfix for initializing non-existant grudge list
		if ((TargetLogState.GuildShare == "true") and (GetGuildInfo("player")) and (not lumpdata)) then
			local sharestr = TargetLog_GuildShareGrudge(name);
			SendChatMessage(sharestr,"GUILD");
		end
	end
end

function TargetLog_RemoveGrudge(arg1)
	local capname;
	local name = string.lower(arg1);

	if TargetLogGrudge[RealmName][name] then
		capname = TargetLogGrudge[RealmName][name].Name;
		TargetLogGrudge[RealmName][name] = nil;
		DEFAULT_CHAT_FRAME:AddMessage("|cffff1f1f"..capname.." removed from grudge list.");
		TotalGrudges = TotalGrudges - 1;

		-- Update dialog box
		TargetLogGrudgeFrame.ListSelected = -1;
		GrudgePanelSelect = nil;
		TargetLogGrudge_Sort(CurrentSortType);
		TargetLogGrudge_UpdateList();
	else
		DEFAULT_CHAT_FRAME:AddMessage("|cffff1f1f"..name.." not found on the grudge list.");
	end
end

function TargetLog_GuildShareGrudge(name)
	local sharestr = "";

	name = string.lower(name);
	sharestr = string.format(GRUDGESHARE_HEADER.." N<%s> L<%s> R<%s> C<%s> A<%s>",
	                         TargetLogGrudge[RealmName][name].Name,
	                         TargetLogGrudge[RealmName][name].LastLevel,
	                         TargetLogGrudge[RealmName][name].Race,
	                         TargetLogGrudge[RealmName][name].Class,
	                         TargetLogGrudge[RealmName][name].Affiliation);
	if (TargetLogGrudge[RealmName][name].LastGuild) then
		sharestr = sharestr.." G<"..TargetLogGrudge[RealmName][name].LastGuild..">";
	end

	return sharestr;
end

function TargetLog_CheckAgainstGrudge(checktype)
	local name, isenemy, key;

	-- If it's a non-player named the same as grudge (eg. a pet), we should ignore.
	if (not UnitIsPlayer(checktype)) then
		return;
	end

	key = string.lower(UnitName(checktype));
	if TargetLogGrudge[RealmName][key] then
		-- Name, level, class and race should be in TargetLogData[1] already, unless FRIENDLY and ONLYHOSTILE set
		if ((TargetLogState.OnlyHostile == "true") and (not UnitIsEnemy(checktype,"player"))) then
			-- In this case, we won't have the TargetLog data to fall back on.
			TargetLogGrudge[RealmName][key].Name = UnitName(checktype);
			TargetLogGrudge[RealmName][key].LastLevel = UnitLevel(checktype);
			TargetLogGrudge[RealmName][key].Race = UnitRace(checktype);
			TargetLogGrudge[RealmName][key].Class = UnitClass(checktype);
			TargetLogGrudge[RealmName][key].Affiliation = FRIENDLY_PC; -- We know it's a friendly already.
		else
			TargetLogGrudge[RealmName][key].Name = TargetLogData[1].Name;
			TargetLogGrudge[RealmName][key].LastLevel = TargetLogData[1].Level;
			TargetLogGrudge[RealmName][key].Race = TargetLogData[1].Type;
			TargetLogGrudge[RealmName][key].Class = TargetLogData[1].Class;
			TargetLogGrudge[RealmName][key].Affiliation = TargetLogData[1].Affiliation;
		end
		TargetLogGrudge[RealmName][key].LastGuild = GetGuildInfo(checktype);
		TargetLogGrudge[RealmName][key].LastSeen = date();

		TargetLogGrudge[RealmName][key].RankName = TargetLogData[1].RankName;
		TargetLogGrudge[RealmName][key].RankNumber = TargetLogData[1].RankNumber;

		if (key ~= CurrentGrudge) then
		-- Don't spam 2nd warning if it's the same as the current one
			TargetLog_GrudgeWarning(key)
			CurrentGrudge = key;
		end
		return "true";
	end
	return "false";
end

function TargetLog_GrudgeWarning(key)
	local warning = GRUDGEWARNCOLOUR.."ALERT: ";
	local firstline = "";

	-- Generate a grudge warning
	-- Also update our grudge data eg. if player has increased in level or has changed guild
	warning = warning..TargetLog_GetNameColour(TargetLogGrudge[RealmName][key].Affiliation)..TargetLogGrudge[RealmName][key].Name..GRUDGEWARNCOLOUR..", ";
	warning = warning..TargetLogGrudge[RealmName][key].LastLevel.." ";
	warning = warning..TargetLogGrudge[RealmName][key].Race.." ";
	warning = warning..TargetLogGrudge[RealmName][key].Class;
	if TargetLogGrudge[RealmName][key].LastGuild then
		warning = warning.." <"..TargetLogGrudge[RealmName][key].LastGuild..">";
	end
	-- Rank
	if (TargetLogGrudge[RealmName][key].RankNumber > 0) then
		warning = warning.." "..TargetLogGrudge[RealmName][key].RankName.."("..TargetLogGrudge[RealmName][key].RankNumber..")";
	end

	-- Tag on first line of 'reason'
	if TargetLogGrudge[RealmName][key].Reason then
		firstline = string.gsub(TargetLogGrudge[RealmName][key].Reason,"\n.*$","");
		warning = warning.."\n"..firstline;	
	end
	UIErrorsFrame:AddMessage(warning, 1.0, 1.0, 1.0, 1.0, UIERRORS_HOLD_TIME);
	-- Show it in our chat frame as well so can scroll back to see
	DEFAULT_CHAT_FRAME:AddMessage(warning);
	if (not TargetLogState.NoSound) or (TargetLogState.NoSound ~= "true") then
		PlaySound("AuctionWindowOpen");
	end
end

function TargetLog_ParseGuildMsg(msg, player)
	local i, j;
	local name, level, race, class, affiliation, guild;	
	local lumpdata;
	--local guildstr = "";

	--DEFAULT_CHAT_FRAME:AddMessage("Received: "..msg.." from "..player);
	if (player == UnitName("player")) then
		-- From self;
		return;
	end
	if (string.find(msg,GRUDGESHARE_HEADER) == 1) then
		--DEFAULT_CHAT_FRAME:AddMessage("Received: "..msg.." from "..player);
		i,j, name,level,race,class,affiliation =
			string.find(msg, GRUDGESHARE_HEADER.." N<([^>]+)> L<([^>]+)> R<([^>]+)> C<([^>]+)> A<([^>]+)>");

		if (name and level and race and class and affiliation) then
			if (tonumber(level)) then
				level = tonumber(level);
			else
				level = "Unknown Level";
			end
			lumpdata = {};
			lumpdata.Name = name;
			lumpdata.LastLevel = level;
			lumpdata.Race = race;
			lumpdata.Class = class;
			lumpdata.Affiliation = affiliation;
			-- For reason, we will say: "<date> From <player>"
			lumpdata.Reason = date()..": From "..player;

			-- See if guild data is there
			guild = nil;
			i,j, guild = string.find(msg, GRUDGESHARE_HEADER..".*G<([^>]+)>");
			if (guild) then
				lumpdata.LastGuild = guild;
				--guildstr=" <"..guild..">"
			end

			--DEFAULT_CHAT_FRAME:AddMessage("Received grudge: "..name.." "..level.." "..race.." "..class..guildstr);
			TargetLog_AddGrudge(name,"false", lumpdata);
		else
			--DEFAULT_CHAT_FRAME:AddMessage("Error parsing");
		end
	end
end

--==========================================================================--
-- This is code for the Grudge List window

function TargetLogGrudge_Init()
	GrudgePanelSelect = nil;
	TargetLogGrudgeTitleText:SetText("Grudge List ("..GetCVar("realmName")..")");
end

function TargetLogGrudge_Open()
	PlaySound("igMainMenuOpen");
	TargetLogGrudge_Sort(CurrentSortType);
	TargetLogGrudge_UpdateList();
	TargetLogGrudge_UpdateButtonStates();
end

function TargetLogGrudge_Close()
	PlaySound("igMainMenuClose");
	-- FIXME:
	TargetLogGrudgeFrame.ListSelected = -1
	GrudgePanelSelect = nil;
	--TargetLogGrudge_Sort(CurrentSortType);
	--TargetLogGrudge_UpdateList();
end

function TargetLogGrudge_Toggle()
	if (TargetLogGrudgeFrame:IsVisible()) then
		HideUIPanel(TargetLogGrudgeFrame);
	else
		ShowUIPanel(TargetLogGrudgeFrame);
	end
end

function TargetLogGrudge_ListButton()
	local ColouredName;
	local id = this:GetID();
	local cltime = GetTime();

	-- Figure out who's name we're on
	TargetLogGrudgeFrame.ListSelected = getglobal("TargetLogGrudgeButton"..id).ListIndex;
	-- Name probably has a colour header on it, so strip it off
	--ColouredName = getglobal("TargetLogGrudgeButton"..this:GetID().."Name"):GetText();
	--GrudgePanelSelect = string.gsub(ColouredName,"^c|........","");
	GrudgePanelSelect = SortedGrudgeList[TargetLogGrudgeFrame.ListSelected].Name;
	TargetLogGrudgeFrame.NameSelected = GrudgePanelSelect;

	-- Handle shift-click;
	if (IsShiftKeyDown() and ChatFrameEditBox:IsVisible()) then
		local sharestr = TargetLog_GuildShareGrudge(GrudgePanelSelect);
		ChatFrameEditBox:Insert(sharestr);
		return;
	end

	-- Update the window
	TargetLogGrudge_UpdateList();
	TargetLogGrudge_UpdateButtonStates();

	-- Handle double-click;
	if ((DoubleClick[1] == id) and ((cltime - DoubleClick[2]) < TLOG_DOUBLECLICK_TIME)) then
		TargetLogGrudgeReason_BeginEdit(GrudgePanelSelect);
	end
	DoubleClick = {id, cltime};
end

function TargetLogGrudge_ReasonButton()
	TargetLogGrudge_ListButton();

	local name = GrudgePanelSelect;
end

function TargetLogGrudge_ButtonOnEnter()
	local id = this:GetID();
	local index = getglobal("TargetLogGrudgeButton"..id).ListIndex;
	local key = string.lower(SortedGrudgeList[index].Name);
	local name = TargetLogGrudge[RealmName][key].Name;
	local reason = TargetLogGrudge[RealmName][key].Reason;
	local guild;
	local lastseen;
	local rank;
	
	if (TargetLogGrudge[RealmName][key].RankNumber > 0) then
		rank = TargetLogGrudge[RealmName][key].RankName.."("..TargetLogGrudge[RealmName][key].RankNumber..")";
	else
		rank = nil; 
	end

	if (TargetLogGrudge[RealmName][key].LastSeen) then
		lastseen = "Last seen: "..TargetLogGrudge[RealmName][key].LastSeen;
	else
		lastseen = "Last seen: Unknown";
	end

	if (TargetLogGrudge[RealmName][key].LastGuild) then
		guild = " <"..TargetLogGrudge[RealmName][key].LastGuild..">";
	else
		guild = "";
	end

	-- Set the tooltip text.
	if (reason == "") then
		reason = "<none set>";
	end

	-- Option to add PvP data later 

	GameTooltip:SetOwner(getglobal("TargetLogGrudgeButton"..id),"ANCHOR_BOTTOMRIGHT");
	--GameTooltip:SetText("Reason:",1.0,1.0,1.0,1);
	--GameTooltip:SetText(name..guild,1.0,1.0,1.0,1);
	GameTooltip:AddDoubleLine(name,guild,1.0,1.0,1.0, 1.0,0.75,0.0);
	if rank then GameTooltip:AddLine(rank,0.75,0.75,0.75,1); end
	GameTooltip:AddLine(reason,1.0,1.0,0.25,1);
	GameTooltip:AddDoubleLine("(double-click to edit reason)", lastseen,0.5,0.5,0.5,1);
	GameTooltip:Show();
end

function TargetLogGrudge_UpdateList()
	local TotalText = "";
	local ListOffset = FauxScrollFrame_GetOffset(TargetLogGrudgeScrollFrame);
	local ListIndex;
	local ListButton;
	local GuildButtonArea;
	local ShowScrollBar = nil;
	local AffColour = GRUDGEFRAMECOLOUR;
	local NameText, LevelText, RaceText, ClassText, GuildText;

	-- summation
	if (TotalGrudges == 1) then
		TotalText = "1 Player Found";
	else
		TotalText = TotalGrudges.." Players Found";
	end
	TargetLogGrudgeTotalText:SetText(TotalText);

	-- Guild Share toggle
	if (GetGuildInfo("player")) then
		if (TargetLogState.GuildShare == "true") then
			TargetLogGrudgeGuildShareToggle:SetChecked(1);
		else
			TargetLogGrudgeGuildShareToggle:SetChecked(0);
		end
	else
		TargetLogGrudgeGuildShareToggle:SetChecked(0);
		TargetLogGrudgeGuildShareToggle:Disable();
	end

	-- Possible substitutions if necessary:
	--	Night Elf	N.Elf
	--	Undead		Und.
	--	Tauren		Taur.

	--	Warrior		Warr.
	--	Paladin		Pala.
	--	Shaman		Sham.
	--	Hunter		Hunt.
	--	Warlock		WrLk.

	-- Create the display list
	if (TotalGrudges > GRUDGEFRAME_DISPLAYNUM) then
		ShowScrollBar = 1;
	end
	for i=1, GRUDGEFRAME_DISPLAYNUM, 1 do
		ListIndex = ListOffset + i;
		ListButton = getglobal("TargetLogGrudgeButton"..i);
		ListButton.ListIndex = ListIndex;

		if (ListIndex <= TotalGrudges) then
			AffColour = TargetLog_GetNameColour(SortedGrudgeList[ListIndex].Affiliation);
			NameText = SortedGrudgeList[ListIndex].Name;

			LevelText = SortedGrudgeList[ListIndex].LastLevel;
			if (string.find(LevelText,"Unknown") == 1) then LevelText = "??"; end

			RaceText = SortedGrudgeList[ListIndex].Race;
			-- Substitutions:
			RaceText = string.gsub(RaceText,"Night Elf","N.Elf");
			if (string.find(RaceText,"Unknown") == 1) then RaceText = "??"; end

			ClassText = SortedGrudgeList[ListIndex].Class;
			-- Substitutions:
			--ClassText = string.gsub(ClassText,"Warlock","W'lock");
			if (string.find(ClassText,"Unknown") == 1) then ClassText = "??"; end

			if (SortedGrudgeList[ListIndex].LastGuild) then
				GuildText = SortedGrudgeList[ListIndex].LastGuild;
			else
				GuildText = "";
			end
		else
			AffColour = GRUDGEFRAMECOLOUR;
			NameText = "";
			LevelText = "";
			RaceText = "";
			ClassText = "";
			GuildText = "";
		end
		getglobal("TargetLogGrudgeButton"..i.."Name"):SetText(AffColour..NameText);
		getglobal("TargetLogGrudgeButton"..i.."Level"):SetText(GRUDGEFRAMECOLOUR..LevelText);
		getglobal("TargetLogGrudgeButton"..i.."Race"):SetText(GRUDGEFRAMECOLOUR..RaceText);
		getglobal("TargetLogGrudgeButton"..i.."Class"):SetText(GRUDGEFRAMECOLOUR..ClassText);
		GuildButton = getglobal("TargetLogGrudgeButton"..i.."Guild");
		GuildButton:SetText(AffColour..GuildText);


		if (ShowScrollBar) then
			GuildButton:SetWidth(78);
		else
			GuildButton:SetWidth(98);
		end
		
		-- Highlights
		if (TargetLogGrudgeFrame.ListSelected == ListIndex) then
			ListButton:LockHighlight();
		else
			ListButton:UnlockHighlight();
		end
		if (ListIndex > TotalGrudges) then
			ListButton:Hide();
		else
			ListButton:Show();
		end
	end

	-- No scroll bar ... can enlarge last tab a bit.
	if not ShowScrollBar then
		TargetLogGrudge_ColumnSetWidth(107,TargetLogGrudgeColumnHeader5);

	else
		TargetLogGrudge_ColumnSetWidth(92,TargetLogGrudgeColumnHeader5);
	end

	FauxScrollFrame_Update(TargetLogGrudgeScrollFrame, TotalGrudges, GRUDGEFRAME_DISPLAYNUM, GRUDGEFRAME_HEIGHT);
end

function TargetLogGrudge_RemoveEntry()
	if (not GrudgePanelSelect) then
		-- Shouldn't be able to do that, since the button should be disabled
		return;
	else
		TargetLog_RemoveGrudge(GrudgePanelSelect);
		TargetLogGrudge_Sort(CurrentSortType);
		TargetLogGrudge_UpdateList();
		TargetLogGrudge_UpdateButtonStates();
	end
end

function TargetLogGrudge_UpdateButtonStates()
	if (GrudgePanelSelect) then
		TargetLogGrudgeRemoveButton:Enable();
	else
		TargetLogGrudgeRemoveButton:Disable();
	end
	if (UnitIsPlayer("target") and (UnitName("target") ~= UnitName("player")))then
		TargetLogGrudgeAddButton:Enable();
	else
		TargetLogGrudgeAddButton:Disable();
	end
end

function TargetLogGrudge_ColumnSetWidth(width, frame)
	if (not frame) then
		frame = this;
	end
	frame:SetWidth(width);
	getglobal(frame:GetName().."Middle"):SetWidth(width - 9);
end

function TargetLogGrudge_Sort(sortby, clicked)
	local i, j, k;
	local EnemyGrudgeList = { };
	local FriendGrudgeList = { };
	local UnknownGrudgeList = { };

	local Reversed = (clicked and (CurrentSortType == sortby) and (not SortReversed));
	local LocalSortOrder = DEFAULT_SORT_ORDER;
	local FuncOrder = { nil, nil, nil, nil, nil };

	-- Generate number-indexed list
	i = 1; j = 1; k = 1;
	for key,value in TargetLogGrudge[RealmName] do
		if (TargetLogGrudge[RealmName][key].Affiliation == ENEMY_PC) then
			EnemyGrudgeList[i] = { };
			EnemyGrudgeList[i].Name = TargetLogGrudge[RealmName][key].Name;
			EnemyGrudgeList[i].Race = TargetLogGrudge[RealmName][key].Race;
			EnemyGrudgeList[i].Class = TargetLogGrudge[RealmName][key].Class;
			EnemyGrudgeList[i].Affiliation = ENEMY_PC;
			EnemyGrudgeList[i].LastLevel = TargetLogGrudge[RealmName][key].LastLevel;
			EnemyGrudgeList[i].LastGuild = TargetLogGrudge[RealmName][key].LastGuild;
			i = i + 1;
		elseif ((TargetLogGrudge[RealmName][key].Affiliation == UNKNOWN_AFF) or
		        (TargetLogGrudge[RealmName][key].Affiliation == "")) then
			UnknownGrudgeList[j] = { };
			UnknownGrudgeList[j].Name = TargetLogGrudge[RealmName][key].Name;
			UnknownGrudgeList[j].Race = TargetLogGrudge[RealmName][key].Race;
			UnknownGrudgeList[j].Class = TargetLogGrudge[RealmName][key].Class;
			UnknownGrudgeList[j].Affiliation = UNKNOWN_AFF;
			UnknownGrudgeList[j].LastLevel = TargetLogGrudge[RealmName][key].LastLevel;
			UnknownGrudgeList[j].LastGuild = TargetLogGrudge[RealmName][key].LastGuild;
			j = j + 1;
		else
			FriendGrudgeList[k] = { };
			FriendGrudgeList[k].Name = TargetLogGrudge[RealmName][key].Name;
			FriendGrudgeList[k].Race = TargetLogGrudge[RealmName][key].Race;
			FriendGrudgeList[k].Class = TargetLogGrudge[RealmName][key].Class;
			FriendGrudgeList[k].Affiliation = FRIENDLY_PC;
			-- We set this to FRIENDLY_PC here, since they could
			-- be a guild or party member, which would end up a
			-- different colour and we want a consistent colour
			FriendGrudgeList[k].LastLevel = TargetLogGrudge[RealmName][key].LastLevel;
			FriendGrudgeList[k].LastGuild = TargetLogGrudge[RealmName][key].LastGuild;
			k = k + 1;
		end
	end

	-- Bump the current sort method to the top;
	i = 1;
	while ((LocalSortOrder[i] ~= sortby) and (i < 5)) do
		i = i + 1;
	end
	table.remove(LocalSortOrder,i);
	table.insert(LocalSortOrder,1,sortby);

	CurrentSortType = sortby;

	-- Sort individual lists
	-- Figure out which order we are sorting in
	for i=1,5,1 do
		if (LocalSortOrder[i] == "name") then FuncOrder[i]="TargetLogCompare_Name";
		elseif (LocalSortOrder[i] == "level") then FuncOrder[i]="TargetLogCompare_Level";
		elseif (LocalSortOrder[i] == "race") then FuncOrder[i]="TargetLogCompare_Race";
		elseif (LocalSortOrder[i] == "class") then FuncOrder[i]="TargetLogCompare_Class";
		elseif (LocalSortOrder[i] == "guild") then FuncOrder[i]="TargetLogCompare_Guild";
		else
			FuncOrder[i]="TargetLogCompare_Name";	-- fallback in case of error;
		end
	end
	if (Reversed) then
		FuncOrder[1] = FuncOrder[1].."Reversed";
	end

	-- Do the sorting of each individual list
	--for i=1,5,1 do
	--	DEFAULT_CHAT_FRAME:AddMessage("FuncOrder["..i.."] = "..FuncOrder[i]);
	--end

	for i=5,1,-1 do
		--DEFAULT_CHAT_FRAME:AddMessage("Sorting #"..i.." with "..FuncOrder[i].."()");
		table.sort(EnemyGrudgeList, getglobal(FuncOrder[i]));
		table.sort(UnknownGrudgeList, getglobal(FuncOrder[i]));
		table.sort(FriendGrudgeList, getglobal(FuncOrder[i]));
	end

	--table.sort(EnemyGrudgeList, TargetLogCompare_Name);
	--table.sort(UnknownGrudgeList, TargetLogCompare_Name);
	--table.sort(FriendGrudgeList, TargetLogCompare_Name);

	-- Concatenate them: Hostiles first, then Unknowns, then Friendlies
	i = 1;
	SortedGrudgeList = { };
	for j = 1,table.getn(EnemyGrudgeList),1 do
		SortedGrudgeList[i] = { };
		SortedGrudgeList[i].Name = EnemyGrudgeList[j].Name;
		SortedGrudgeList[i].Race = EnemyGrudgeList[j].Race;
		SortedGrudgeList[i].Class = EnemyGrudgeList[j].Class;
		SortedGrudgeList[i].Affiliation = EnemyGrudgeList[j].Affiliation;
		SortedGrudgeList[i].LastLevel = EnemyGrudgeList[j].LastLevel;
		SortedGrudgeList[i].LastGuild = EnemyGrudgeList[j].LastGuild;
		i = i + 1;
	end
	for j = 1,table.getn(UnknownGrudgeList),1 do
		SortedGrudgeList[i] = { };
		SortedGrudgeList[i].Name = UnknownGrudgeList[j].Name;
		SortedGrudgeList[i].Race = UnknownGrudgeList[j].Race;
		SortedGrudgeList[i].Class = UnknownGrudgeList[j].Class;
		SortedGrudgeList[i].Affiliation = UnknownGrudgeList[j].Affiliation;
		SortedGrudgeList[i].LastLevel = UnknownGrudgeList[j].LastLevel;
		SortedGrudgeList[i].LastGuild = UnknownGrudgeList[j].LastGuild;
		i = i + 1;
	end
	for j = 1,table.getn(FriendGrudgeList),1 do
		SortedGrudgeList[i] = { };
		SortedGrudgeList[i].Name = FriendGrudgeList[j].Name;
		SortedGrudgeList[i].Race = FriendGrudgeList[j].Race;
		SortedGrudgeList[i].Class = FriendGrudgeList[j].Class;
		SortedGrudgeList[i].Affiliation = FriendGrudgeList[j].Affiliation;
		SortedGrudgeList[i].LastLevel = FriendGrudgeList[j].LastLevel;
		SortedGrudgeList[i].LastGuild = FriendGrudgeList[j].LastGuild;
		i = i + 1;
	end
	TotalGrudges = i - 1;

	SortReversed = Reversed;
end

------------------------------------------------------------------------------
-- And here we go for sort comparison functions... yay...

function TargetLogCompare_Name(a, b)		return a.Name < b.Name; end
function TargetLogCompare_NameReversed(a, b)	return a.Name > b.Name; end

-----

--function TargetLogCompare_Level(a, b)		return tonumber(a.LastLevel) < tonumber(b.LastLevel); end
function TargetLogCompare_Level(a, b)
	local anum = tonumber(a.LastLevel);
	local bnum = tonumber(b.LastLevel);
	if not anum then anum = 0; end
	if not bnum then bnum = 0; end
	return anum < bnum;
end

--function TargetLogCompare_LevelReversed(a, b)	return tonumber(a.LastLevel) > tonumber(b.LastLevel); end
function TargetLogCompare_LevelReversed(a, b)
	local anum = tonumber(a.LastLevel);
	local bnum = tonumber(b.LastLevel);
	if not anum then anum = 0; end
	if not bnum then bnum = 0; end
	return anum > bnum;
end

-----

function TargetLogCompare_Race(a, b)		return a.Race < b.Race; end
function TargetLogCompare_RaceReversed(a, b)	return a.Race > b.Race; end

-----

function TargetLogCompare_Class(a, b)		return a.Class < b.Class; end
function TargetLogCompare_ClassReversed(a, b)	return a.Class > b.Class; end

-----

function TargetLogCompare_Guild(a, b)
	if (a.LastGuild and b.LastGuild) then
		return a.LastGuild < b.LastGuild;
	else
		if (b.LastGuild) then return 1; else return nil; end
	end
end
function TargetLogCompare_GuildReversed(a, b)
	if (a.LastGuild and b.LastGuild) then
		return a.LastGuild > b.LastGuild;
	else
		if (a.LastGuild) then return 1; else return nil; end
	end
end

--==========================================================================--
-- This is code for the Grudge List Reasong editing window

function TargetLogGrudgeReason_BeginEdit(namestr)
	local name = string.lower(namestr);

	if not TargetLogGrudge[RealmName][name] then
		return;
	end

	IsEditingReason = name;
	TargetLogGrudgeReasonFrameHeaderString:SetText(TargetLogGrudge[RealmName][name].Name);
	TargetLogGrudgeReasonFrameEditBox:SetText(TargetLogGrudge[RealmName][name].Reason);
	TargetLogGrudgeReasonFrame:Show();
end

function TargetLogGrudgeReason_Save()
	local reasontext;

	if (not IsEditingReason) then
		return;
	end

	reasontext = TargetLogGrudgeReasonFrameEditBox:GetText();
	TargetLogGrudge[RealmName][IsEditingReason].Reason = reasontext;

	IsEditingReason = nil;
	TargetLogGrudgeReasonFrameEditBox:SetText("");
	TargetLogGrudgeReasonFrame:Hide();
end

function TargetLogGrudgeReason_Cancel()
	IsEditingReason = nil;
	TargetLogGrudgeReasonFrame:Hide();
end
